from typing import Any, Callable, Dict, Generator, List, TypeVar, Union

from graia.broadcast.utilles import NestableIterable

from ..typing import DEFAULT_LIFECYCLE_NAMES, T_Dispatcher

T = TypeVar("T")
I = TypeVar("I")

LF_TEMPLATE = {i: list() for i in DEFAULT_LIFECYCLE_NAMES}


class ExecutionContext:
    __slots__ = ("event", "lifecycle_refs", "dispatchers", "path")

    lifecycle_refs: Dict[str, List[Callable]]
    dispatchers: List[T_Dispatcher]

    path: NestableIterable

    def __init__(self, dispatchers: List[T_Dispatcher], path: NestableIterable):
        self.dispatchers = dispatchers
        self.path = path

        self.lifecycle_refs = LF_TEMPLATE.copy()


class ParameterContext:
    __slots__ = ("name", "annotation", "default")

    name: str
    annotation: Any
    default: Any

    def __init__(self, name, annotation, default):
        self.name = name
        self.annotation = annotation
        self.default = default

    def __repr__(self) -> str:
        return "<ParameterContext name={0} annotation={1} default={2}".format(
            self.name, self.annotation, self.default
        )



from typing import TYPE_CHECKING, Any, Callable, Type, Union

if TYPE_CHECKING:
    from graia.broadcast.entities.dispatcher import BaseDispatcher
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface


T_Dispatcher = Union[
    Type["BaseDispatcher"], "BaseDispatcher", Callable[["DispatcherInterface"], Any]
]

T_Dispatcher_Callable = Callable[["DispatcherInterface"], Any]

DEFAULT_LIFECYCLE_NAMES = (
    "beforeExecution",
    "afterDispatch",
    "afterExecution",
)



import typing
from typing import Any, Callable, Optional

from ..entities.decorator import Decorator
from ..entities.exectarget import ExecTarget
from ..entities.signatures import Force
from ..exceptions import RequirementCrashed
from ..interfaces.decorator import DecoratorInterface
from ..utilles import dispatcher_mixin_handler


class Depend(Decorator):
    pre = True
    depend_callable: ExecTarget
    cache: bool = False

    def __init__(self, callable: Callable, *, cache=False):
        self.cache = cache
        self.depend_callable = ExecTarget(callable)

    def __repr__(self) -> str:
        return "<Depend target={0}>".format(self.depend_callable)

    async def target(self, interface: DecoratorInterface):
        if self.cache:
            attempt = interface.local_storage.get(self.depend_callable)
            if attempt:
                return Force(attempt)
        result = await interface.dispatcher_interface.broadcast.Executor(
            target=self.depend_callable,
            dispatchers=interface.dispatcher_interface.execution_contexts[
                -1
            ].dispatchers,
        )

        if self.cache:
            interface.local_storage[self.depend_callable] = result
        return Force(result)


class OptionalParam(Decorator):
    pre = True

    def __init__(self, origin: Any):
        self.origin = origin

    async def target(self, interface: DecoratorInterface) -> Optional[Any]:
        try:
            return Force(
                await interface.dispatcher_interface.lookup_param(
                    interface.dispatcher_interface.name,
                    interface.dispatcher_interface.annotation.__args__[0]
                    if isinstance(interface.dispatcher_interface.annotation, typing._GenericAlias)  # type: ignore
                    and type(None) in interface.dispatcher_interface.annotation.__args__
                    else interface.dispatcher_interface.annotation,
                    self.origin,
                )
            )
        except RequirementCrashed:
            return Force()



from functools import lru_cache
from inspect import getattr_static
from typing import TYPE_CHECKING, Any, Dict, Generic, Iterable, List, TypeVar

from ..entities.context import ExecutionContext, ParameterContext
from ..entities.dispatcher import BaseDispatcher
from ..entities.event import Dispatchable
from ..entities.signatures import Force
from ..exceptions import RequirementCrashed
from ..typing import DEFAULT_LIFECYCLE_NAMES, T_Dispatcher
from ..utilles import NestableIterable, run_always_await_safely

if TYPE_CHECKING:
    from .. import Broadcast


class EmptyEvent(Dispatchable):
    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(_):
            pass


T_Event = TypeVar("T_Event", bound=Dispatchable)

LIFECYCLE_ABS = {
    BaseDispatcher.beforeExecution,
    BaseDispatcher.afterDispatch,
    BaseDispatcher.afterExecution,
}


@lru_cache(maxsize=None)
def get_lifecycle_refs(dispatcher: T_Dispatcher) -> Dict:
    result = {}

    for name in DEFAULT_LIFECYCLE_NAMES:
        d = getattr_static(dispatcher, name)
        h = getattr(d, "__func__", None)
        if h is not None and h not in LIFECYCLE_ABS:
            result[name] = getattr(dispatcher, name)
    return result


class DispatcherInterface(Generic[T_Event]):
    broadcast: "Broadcast"

    execution_contexts: List[ExecutionContext]
    parameter_contexts: List[ParameterContext]

    def __init__(self, broadcast_instance: "Broadcast") -> None:
        self.broadcast = broadcast_instance
        self.execution_contexts = []
        self.parameter_contexts = []

    @property
    def name(self) -> str:
        return self.parameter_contexts[-1].name

    @property
    def annotation(self) -> Any:
        return self.parameter_contexts[-1].annotation

    @property
    def default(self) -> Any:
        return self.parameter_contexts[-1].default

    @property
    def event(self) -> T_Event:
        return self.broadcast.event_ctx.get()  # type: ignore

    @property
    def global_dispatcher(self) -> List[T_Dispatcher]:
        return self.broadcast.global_dispatchers

    def start_execution(
        self,
        dispatchers: List[T_Dispatcher],
    ):
        self.execution_contexts.append(
            ExecutionContext(
                dispatchers, NestableIterable(self.execution_contexts[-1].dispatchers)
            )
        )
        self.flush_lifecycle_refs(self.execution_contexts[-1].dispatchers)
        return self

    def clean(self):
        self.execution_contexts.pop()

    def inject_execution_raw(self, *dispatchers: T_Dispatcher):
        for dispatcher in dispatchers:
            self.execution_contexts[-1].dispatchers.insert(0, dispatcher)

        self.flush_lifecycle_refs(dispatchers)

    def inject_global_raw(self, *dispatchers: T_Dispatcher):
        for dispatcher in dispatchers[::-1]:
            self.broadcast.global_dispatchers.insert(1, dispatcher)

    async def exec_lifecycle(self, lifecycle_name: str, *args, **kwargs):
        lifecycle_funcs = self.execution_contexts[-1].lifecycle_refs.get(lifecycle_name)
        if lifecycle_funcs:
            for func in lifecycle_funcs:
                await run_always_await_safely(func, self, *args, **kwargs)

    def flush_lifecycle_refs(
        self,
        dispatchers: Iterable[T_Dispatcher],
    ):
        lifecycle_refs = self.execution_contexts[-1].lifecycle_refs

        for dispatcher in dispatchers:
            if dispatcher.__class__ is not type and not isinstance(
                dispatcher, BaseDispatcher
            ):
                continue

            result = get_lifecycle_refs(dispatcher)
            for k, v in result.items():
                lifecycle_refs[k].append(v)

    async def lookup_param(
        self,
        name: str,
        annotation: Any,
        default: Any,
        oplog: Any,
    ) -> Any:
        continued = 0
        self.parameter_contexts.append(ParameterContext(name, annotation, default))

        try:
            self.execution_contexts[-1].path = NestableIterable(oplog[0])
            for dispatcher in self.execution_contexts[-1].path:
                result = await getattr(dispatcher, "catch", dispatcher)(self)
                if result is None:  # 不可靠.
                    break
                if result.__class__ is Force:
                    return result.target
                return result
            oplog[0].clear()
            self.execution_contexts[-1].path = self.execution_contexts[-1].path
            for dispatcher in self.execution_contexts[-1].path:
                result = await getattr(dispatcher, "catch", dispatcher)(self)

                if result is None:
                    continued += 1
                    continue

                if result.__class__ is Force:
                    return result.target

                oplog[0].append(dispatcher)
                oplog[1] += continued
                return result
            else:
                raise RequirementCrashed(
                    "the dispatching requirement crashed: ",
                    self.name,
                    self.annotation,
                    self.default,
                )
        finally:
            self.parameter_contexts.pop()

    async def lookup_by_directly(
        self, dispatcher: T_Dispatcher, name: str, annotation: Any, default: Any
    ) -> Any:
        self.parameter_contexts.append(
            ParameterContext(
                name,
                annotation,
                default,
            )
        )

        try:
            result = await getattr(dispatcher, "catch", dispatcher)(self)
            if result.__class__ is Force:
                return result.target

            return result
        finally:
            self.parameter_contexts.pop()



import inspect
from contextlib import contextmanager
from contextvars import ContextVar, Token
from functools import lru_cache
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    List,
    Type,
    TypeVar,
    Union,
)

from .entities.dispatcher import BaseDispatcher


async def run_always_await(any_callable: Union[Awaitable, Callable]):
    if inspect.isawaitable(any_callable):
        return await any_callable
    else:
        return any_callable


async def run_always_await_safely(callable, *args, **kwargs):
    if iscoroutinefunction(callable):
        return await callable(*args, **kwargs)
    return callable(*args, **kwargs)


T = TypeVar("T")
D = TypeVar("D")


class Ctx(Generic[T]):
    current_ctx: ContextVar[T]

    def __init__(self, name: str) -> None:
        self.current_ctx = ContextVar(name)

    def get(self, default: Union[D, T] = None) -> Union[T, D]:
        return self.current_ctx.get(default)

    def set(self, value: T):
        return self.current_ctx.set(value)

    def reset(self, token: Token):
        return self.current_ctx.reset(token)

    @contextmanager
    def use(self, value: T):
        token = self.set(value)
        yield
        self.reset(token)


def printer(value: Any):
    print(value)
    return value


def group_dict(iterable: Iterable, key_callable: Callable[[Any], Any]):
    temp = {}
    for i in iterable:
        k = key_callable(i)
        temp.setdefault(k, [])
        temp[k].append(i)
    return temp


cache_size = 4096


@lru_cache(cache_size)
def argument_signature(callable_target: Callable):
    return [
        (
            name,
            param.annotation
            if param.annotation is not inspect.Signature.empty
            else None,
            param.default if param.default is not inspect.Signature.empty else None,
        )
        for name, param in inspect.signature(callable_target).parameters.items()
    ]


@lru_cache(cache_size)
def is_asyncgener(o):
    return inspect.isasyncgenfunction(o)


@lru_cache(cache_size)
def iscoroutinefunction(o):
    return inspect.iscoroutinefunction(o)


@lru_cache(cache_size)
def isasyncgen(o):
    return inspect.isasyncgen(o)


@lru_cache(None)
def dispatcher_mixin_handler(dispatcher: Union[Type[BaseDispatcher], BaseDispatcher]):
    unbound_mixin = getattr(dispatcher, "mixin", [])
    result = [dispatcher]

    for i in unbound_mixin:
        if issubclass(i, BaseDispatcher):
            result.extend(dispatcher_mixin_handler(i))
        else:
            result.append(i)
    return result


class NestableIterable(Iterable[T]):
    index_stack: list
    iterable: List[T]

    def __init__(self, iterable: List[T]) -> None:
        self.iterable = iterable
        self.index_stack = [0]

    def __iter__(self):
        index = self.index_stack[-1]
        self.index_stack.append(self.index_stack[-1])

        start_offset = index + index and 1
        try:
            for self.index_stack[-1], content in enumerate(
                self.iterable[start_offset:],
                start=start_offset,
            ):
                yield content
        finally:
            self.index_stack.pop()



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from graia.broadcast.typing import T_Dispatcher


@dataclass
class Namespace:
    name: str
    injected_dispatchers: List["T_Dispatcher"] = field(default_factory=list)

    priority: int = 0
    default: bool = False

    hide: bool = False
    disabled: bool = False



from abc import ABCMeta, abstractmethod
from typing import Any, List, Type

from ..entities.decorator import Decorator
from ..entities.event import Dispatchable
from ..typing import T_Dispatcher


class Waiter(metaclass=ABCMeta):
    listening_events: List[Type[Dispatchable]]
    using_dispatchers: List[T_Dispatcher]
    using_decorators: List[Decorator]
    priority: int
    block_propagation: bool

    @classmethod
    def create(
        cls,
        listening_events: List[Type[Dispatchable]],
        using_dispatchers: List[T_Dispatcher] = None,
        using_decorators: List[Decorator] = None,
        priority: int = 15,  # 默认情况下都是需要高于默认 16 的监听吧...
        block_propagation: bool = False,
    ) -> Type["Waiter"]:
        async def detected_event(self) -> Any:
            pass

        return type(
            "AbstractWaiter",
            (cls,),  # type: ignore
            {
                "listening_events": listening_events,
                "using_dispatchers": using_dispatchers,
                "using_decorators": using_decorators,
                "priority": priority,
                "block_propagation": block_propagation,
                "detected_event": abstractmethod(detected_event),
            },
        )

    async def detected_event(self) -> Any:
        pass

    @classmethod
    def create_using_function(
        cls,
        listening_events: List[Type[Dispatchable]],
        using_dispatchers: List[T_Dispatcher] = None,
        using_decorators: List[Decorator] = None,
        priority: int = 15,  # 默认情况下都是需要高于默认 16 的监听吧...
        block_propagation: bool = False,
    ):
        def wrapper(func):
            return type(
                "SingleWaiter",
                (
                    cls.create(
                        listening_events,
                        using_dispatchers,
                        using_decorators,
                        priority,
                        block_propagation,
                    ),
                ),
                {"detected_event": staticmethod(func)},
            )()

        return wrapper
